// binary file viewer Puting
//
// Copyright © 2023 Yamada Yohei <yamadayohei@gmail.com>
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the “Software”), to
// deal in the Software without restriction, including without limitation the
// rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
// sell copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.

#ifndef UNICODE
#define UNICODE
#endif
#define WINVER 0xA00
#include <windows.h>
#include <wchar.h>
#define B_MAIN TEXT("b_main")

int g_pos;
char *g_mem;
size_t g_memsize;
PTSTR g_filename;
int g_cursor;
int g_win_width  = 800;
int g_win_height = 800;

void
dump(HDC h, char *m)
{
   HFONT h_font = CreateFont(
         /*height=*/20, /*width=*/0, /*escapement=*/0, /*orientation=*/0,
         /*weight=*/FW_NORMAL,
         /*itaric=*/FALSE, /*underline=*/FALSE, /*strike=*/FALSE,
         /*charset=*/ANSI_CHARSET,
         OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY,
         FIXED_PITCH | FF_DONTCARE, TEXT("Consolas"));
   if (h_font) SelectObject(h, h_font);

   TCHAR s[10];
   const int line_height = 24;
   const int n_lines = g_win_height / line_height;

   for (int i = 0; i < n_lines; i++) {
      int x = 0, y = line_height * i;
      swprintf(s, 10, TEXT("%08x"), (i + g_pos) * 16);
      size_t l = lstrlen(s);
      TextOut(h, x + 16, y + 16, s, l);
      for (int j = 0; j < 16; j++) {
         int x = 32 * j;
x += j / 8 * 8;
x += j / 4 * 8;
x += j / 2 * 0;
         int ij = (i + g_pos) * 16 + j;
         unsigned int n = 0;
         if (ij < g_memsize) {
            n = 0xff & (unsigned int)m[ij];
            swprintf(s, 3, TEXT("%02x"), n); }
         else
            swprintf(s, 3, TEXT("--"));
         size_t l = lstrlen(s);
         TextOut(h, x + 120, y + 16, s, l); }
      for (int j = 0; j < 16; j++) {
         int x = 16 * j;
         int ij = (i + g_pos) * 16 + j;
         unsigned int n = 0;
         if (ij < g_memsize)
            n = 0xff & (unsigned int)m[ij];
         if (n < ' ') n = '.';
         if (n >= 0x7f && n < 0xa0) n = '.';
         if (n == ' ') n = 0xa0;
         wchar_t c = (wchar_t)n;
         swprintf(s, 3, TEXT("%C"), c);
         size_t l = lstrlen(s);
         TextOut(h, x + 700, y + 16, s, l); } }

   const int cursor = g_cursor - g_pos * 16;
   if (cursor >= 0 && cursor < n_lines * 16 && g_cursor < g_memsize) {
      int x = cursor % 16 * 32;
      const int j = cursor % 16;
x += j / 8 * 8;
x += j / 4 * 8;
x += j / 2 * 0;
      const int y = cursor / 16 * line_height;
      const unsigned int n = 0xff & (unsigned int)m[g_cursor];
      swprintf(s, 3, TEXT("%02x"), n);
      size_t l = lstrlen(s);
      const COLORREF old_fg = SetTextColor(h, RGB(0xff, 0xff, 0xff));
      const COLORREF old_bg = SetBkColor(h, RGB(0x33, 0x33, 0x66));
      TextOut(h, x + 120, y + 16, s, l);
      SetTextColor(h, old_fg);
      SetBkColor(h, old_bg);
   }

   if (g_cursor >= 0 && g_cursor < g_memsize) {
      TCHAR s[32];
      unsigned char * const p = (unsigned char *)&m[g_cursor];
      unsigned int n = g_cursor;
      swprintf(s, 32, TEXT("cursor: %08x"), n);
      size_t l = lstrlen(s);
      TextOut(h, 1200, 0 * 24 + 16, s, l);

      n = 0xff & *p;
      swprintf(s, 32, TEXT("u8: %02x"), n);
      l = lstrlen(s);
      TextOut(h, 1200, 1 * 24 + 16, s, l);

      n |= (0xff & p[1]) << 8;
      swprintf(s, 32, TEXT("u16le: %04x"), n);
      l = lstrlen(s);
      TextOut(h, 1200, 2 * 24 + 16, s, l);

      n |= (0xff & p[2]) << 16;
      n |= (0xff & p[3]) << 24;
      swprintf(s, 32, TEXT("u32le: %08x"), n);
      l = lstrlen(s);
      TextOut(h, 1200, 3 * 24 + 16, s, l);

      n = (0xff & p[0]) << 8 | 0xff &p[1];
      swprintf(s, 32, TEXT("u16be: %04x"), n);
      l = lstrlen(s);
      TextOut(h, 1200, 4 * 24 + 16, s, l);

      n = n << 16 | (0xff & p[2]) << 8 | 0xff &p[3];
      swprintf(s, 32, TEXT("u32be: %08x"), n);
      l = lstrlen(s);
      TextOut(h, 1200, 5 * 24 + 16, s, l);
   }

   SelectObject(h, GetStockObject(SYSTEM_FONT));
   if (h_font) DeleteObject(h_font);

   HBRUSH brush_red    = CreateSolidBrush(RGB(0xff, 0x00, 0x33));
   HBRUSH brush_grey   = CreateSolidBrush(RGB(0x99, 0x99, 0x99));
   HBRUSH brush_green  = CreateSolidBrush(RGB(0x33, 0xcc, 0x66));
   HBRUSH brush_yellow = CreateSolidBrush(RGB(0xff, 0xcc, 0x33));
   HBRUSH brush_white  = GetStockObject(WHITE_BRUSH);
   HGDIOBJ old_pen   = SelectObject(h, GetStockObject(WHITE_PEN));
   HGDIOBJ old_brush = SelectObject(h, GetStockObject(WHITE_BRUSH));

   const int n_grph_lines = g_win_height / 4;
   const int n_grph_top = g_pos < n_grph_lines / 2 ? n_grph_lines / 2 : g_pos;
   const int x = 1000 - 8;
   const int y = (g_pos - n_grph_top + n_grph_lines / 2) * 4 + 16;
   SelectObject(h, brush_white);
   Rectangle(h, x, 0, x + 4, g_win_height);
   SelectObject(h, brush_yellow);
   Rectangle(h, x, y, x + 4, y + n_lines * 4);

   for (int i = 0; i < n_grph_lines; i++) {
      const int y = 4 * i + 16;
      const int i2 = i - n_grph_lines / 2;

      for (int j = 0; j < 16; j++) {
         const int x = 8 * j + 1000;
         const int ij = (i2 + n_grph_top) * 16 + j;
         const int c = ij >= 0 && ij < g_memsize ? 0xff & m[ij] : 0;

         switch (c) {
         case 0:
            SelectObject(h, brush_white);
            break;
         case 1 ... 0x20:
         case 0x7f:
            SelectObject(h, brush_red);
            break;
         case 0x21 ... 0x7e:
            SelectObject(h, brush_grey);
            break;
         case 0x80 ... 0xa0:
         case 0xff:
            SelectObject(h, brush_green);
            break;
         case 0xa1 ... 0xfe:
            SelectObject(h, brush_yellow);
            break;
         }
         Rectangle(h, x, y, x + 8, y + 4); } }

   SelectObject(h, old_pen);
   SelectObject(h, old_brush);
   DeleteObject(brush_red);
   DeleteObject(brush_green);
   DeleteObject(brush_yellow);
}

#define CTRL(x) (x & 0x1f)

LRESULT CALLBACK
wproc(HWND w, UINT m, WPARAM wp, LPARAM lp)
{
   const int line_height = 24;
   const int n_lines = g_win_height / line_height;

   switch (m) {
   case WM_CHAR:
      InvalidateRect(w, NULL, FALSE);
      switch ((TCHAR)wp) {
      case L'q': // FALLTHRU
      case L'Q': PostQuitMessage(0); break;
      case CTRL('l'): InvalidateRect(w, NULL, TRUE); break;

      case L'j': g_cursor += 16; break;
      case L'k': g_cursor -= 16; break;
      case L'l': g_cursor++; break;
      case L'h': g_cursor--; break;
      case L'L': g_pos = g_cursor / 16; break;
      case L'H': g_cursor = g_pos * 16; break;
      case L'w': g_cursor = (g_cursor / 4 + 1) * 4; break;
      case L'b': g_cursor = (g_cursor - 1) / 4 * 4; break;
      case L'^': g_cursor &= ~0xf; break;
      case L'+': g_cursor = (g_cursor & ~0xf) + 16; break;
      case L'-': g_cursor = (g_cursor & ~0xf) - 16; break;

      case CTRL('j'): // FALLTHRU
      case L'e': g_cursor += 16; g_pos++; break;
      case CTRL('k'): // FALLTHRU
      case L'y': g_cursor -= 16; g_pos--; break;
      case CTRL('f'): g_cursor += n_lines * 16; g_pos += n_lines; break;
      case CTRL('b'): g_cursor -= n_lines * 16; g_pos -= n_lines; break;
      case L'd': g_cursor += n_lines / 2 * 16; g_pos += n_lines / 2; break;
      case L'u': g_cursor -= n_lines / 2 * 16; g_pos -= n_lines / 2; break;
      case L'J': g_pos += g_memsize / 160; g_cursor = g_pos * 16; break;
      case L'K': g_pos -= g_memsize / 160; g_cursor = g_pos * 16; break;
      case L'g': // FALLTHRU
      case L'<': g_cursor = g_pos = 0; break;
      case L'G': // FALLTHRU
      case L'>': g_pos = (g_cursor = g_memsize) / 16; break;
      }
      if (g_pos < 0) g_pos = 0;
      if (g_pos * 16 > g_memsize) g_pos = g_memsize / 16;
      if (g_cursor < 0) g_cursor = 0;
      if (g_cursor > g_memsize) g_cursor = g_memsize;
      break;
   case WM_DESTROY:
      PostQuitMessage(0);
      break;
   case WM_DROPFILES:
      MessageBox(NULL, TEXT("files are dropped"), TEXT("Puting"), MB_OK);
      break;
   case WM_LBUTTONDOWN:
      InvalidateRect(w, NULL, FALSE);
      if (wp & MK_SHIFT) g_pos += n_lines;
      else if (wp & MK_CONTROL) g_pos += g_memsize / 160; else g_pos++;
      if (g_pos * 16 > g_memsize) g_pos = g_memsize / 16;
      break;
   case WM_RBUTTONDOWN:
      InvalidateRect(w, NULL, FALSE);
      if (wp & MK_SHIFT) g_pos -= n_lines;
      else if (wp & MK_CONTROL) g_pos -= g_memsize / 160; else g_pos--;
      if (g_pos < 0) g_pos = 0;
      break;
   case WM_MOUSEWHEEL:
      InvalidateRect(w, NULL, FALSE);
      g_pos -= (short)HIWORD(wp) / (WHEEL_DELTA / 3);
// fprintf(stderr, "(%d %d)\n", (short)HIWORD(wp), (short)HIWORD(wp) / (WHEEL_DELTA / 3));
      if (g_pos < 0) g_pos = 0;
      if (g_pos * 16 > g_memsize) g_pos = g_memsize / 16;
      break;
   case WM_PAINT: {
      PAINTSTRUCT ps;
      const HDC h_dc = BeginPaint(w, &ps);
      dump(h_dc, g_mem);
      EndPaint(w, &ps);
      break; }
   case WM_SIZE:
      g_win_width  = LOWORD(lp);
      g_win_height = HIWORD(lp);
      g_win_height = HIWORD(lp) - 32 /*DEBUG*/;
      break;
   default: return DefWindowProc(w, m, wp, lp);
   }
      return 0;
}

int
register_bmain_wc(HINSTANCE h_inst)
{
   WNDCLASS wc = {
      .hCursor = LoadCursor(NULL, IDC_ARROW),
      .hInstance = h_inst,
      .hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH),
      .lpfnWndProc = wproc,
      .lpszClassName = B_MAIN,
      .style = CS_HREDRAW | CS_VREDRAW,
   };
   return RegisterClass(&wc);
}

PCTSTR
basename(PCTSTR s)
{
   for (PCTSTR i = s; *i; i++)
      if (*i == L'\\') s = i + 1;

   return s;
}

HWND
create_bmain_window(HINSTANCE h_inst)
{
   const int x = 200, y = 200;
   const int w = g_win_width;
   const int h = g_win_height;
   const DWORD style = WS_OVERLAPPEDWINDOW;
   const HWND parent = NULL;
   const HMENU menu = NULL;
   const PVOID lp = NULL;

   TCHAR title[64];
   swprintf(title, 64, TEXT("Puting - %s"), basename(g_filename));

   HWND h_wnd = CreateWindow(B_MAIN,
         title, style, x, y, w, h, parent, menu, h_inst, lp);
   return h_wnd;
}

int
mapping(PCTSTR filename)
{

   HANDLE f = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
         FILE_ATTRIBUTE_NORMAL, NULL);
   if (f == INVALID_HANDLE_VALUE) return 1;

   HANDLE fm = CreateFileMapping(f, NULL, PAGE_READONLY, 0, 0, NULL);
   if (!fm) return 1;

   g_mem = (char *)MapViewOfFile(fm, FILE_MAP_READ, 0, 0, 0);
   if (!g_mem) return 1;

   MEMORY_BASIC_INFORMATION mbi;
   if (!VirtualQuery(g_mem, &mbi, sizeof mbi)) return 1;
   g_memsize = mbi.RegionSize;

   return 0;
}

int WINAPI
wWinMain(
HINSTANCE h_inst,
HINSTANCE h_previnst,
PTSTR     s_cmdline,
int       e_cmdshow)
{
   SetThreadDpiAwarenessContext(DPI_AWARENESS_CONTEXT_SYSTEM_AWARE);

   g_filename = s_cmdline;
   if (mapping(g_filename)) {
      MessageBox(NULL, TEXT("failed to open file"), TEXT("Puting"), MB_OK);
      return 1; }

   if (!register_bmain_wc(h_inst)) return 1;

   HWND h_wnd;
   if ((h_wnd = create_bmain_window(h_inst)) == NULL) return 1;
   ShowWindow(h_wnd, e_cmdshow);
   DragAcceptFiles(h_wnd, TRUE);

   MSG mesg;
   while (GetMessage(&mesg, NULL, 0, 0)) {
      TranslateMessage(&mesg);
      DispatchMessage(&mesg); }
   return mesg.wParam;
}
// vim:sw=3:expandtab:
